/// Josh's User Interface Library
/// A C++20 Library for creating, styling, and rendering of a UI/UX widgets.
/// Developed and Maintained by Josh O'Leary @ Redacted Software
/// Special Thanks to William Tomasine II and Maxine Hayes.
/// (c) 2024 Redacted Software
/// This work is dedicated to the public domain.

/// @file Pinnable.hpp
/// @desc A mixin class that provides a widget with the ability to be "pinned": locked open and fixed in place.
/// @edit 2025-04-30

#pragma once

#include <Event.h>
#include <JUI/UDim2.hpp>

namespace JUI {

    /// The pinnable mixin class. Designates a widget as "pinned" or not, which, depending on the derived classes' semantics, may lock it open or in it's current place.
    class Pinnable {
    public:


        Pinnable() = default;
        virtual ~Pinnable() = default;

        Event<> OnPin;
        Event<> OnUnpin;

        /// Sets the object as "Pinned". Depending on the widget, position and visibility status will be locked.
        /// @see Unpin(), SetPinned().
        virtual void Pin();

        /// Sets the object as "Unpinned".
        /// @see Pin(), SetPinned().
        virtual void Unpin();

        /// Sets the objects pinned status manually.
        void SetPinned(bool pinned);

        /// Called when the pinned status changes, and triggers callback event signals.
        virtual void OnPinStatusChanged(bool new_status);

        bool Pinned() const;

    protected:

        // TODO: Implement position pinning, both relative to parent and global.
        bool pin_point_is_relative = true;
        UDim2 pinned_point_relative{};
        Vector2 pinned_point_global{};
        bool pinned_visibility = false;

    private:
        bool pinned = false;
    };
}