/// Josh's User Interface Library
/// A C++20 Library for creating, styling, and rendering of a UI/UX widgets.
/// Developed and Maintained by Josh O'Leary @ Redacted Software.
/// Special Thanks to William Tomasine II and Maxine Hayes.
/// (c) 2024 Redacted Software
/// This work is dedicated to the public domain.

/// @file Widget.hpp
/// @desc Base Widget Class - All JUI widgets extend their behavior from this class.
/// @edit 2024-10-11

#pragma once

#include <JUI/Widgets/TextButton.hpp>

#include "ListLayout.hpp"
#include "Separator.hpp"
#include "JUI/Mixins/Pinnable.hpp"
#include "../Style/TextStyler.hpp"

namespace JUI {

    class ContextMenu;

    // TODO: Move to NestableMenu.hpp after resolving circular dependency.

    /// An abstract class that defines the interface for nestable menus - menu objects that contain similar sub-menus.
    class NestableMenu {
        virtual ContextMenu* AddSubmenu(const std::string& text) = 0;
        virtual TextButton* AddButton(const std::string& text) = 0;
        virtual TextButton* AddButton(const std::string& text, const std::function<void()>& callback) = 0;
        virtual Separator* AddSeparator(const UDim& size = 5_px) = 0;
    };





    /// A vertically-descending list menu of button objects, which supports sub-menus much like class UtilityBar.
    /// @see UtilityBar
    class ContextMenu  : public Rect, public Pinnable, public NestableMenu, public TextStyler {
    public:
        /// The default constructor initializes base state of the widget.
        ContextMenu();
        /// Constructs this object by explicitly specifying the parent element.
        explicit ContextMenu(Widget* parent);
    public:
        Event<> PresumeMouseFocusLost;
    public:
        /// Adds a new button to this ContextMenu.
        /// @param name The label to apply to the new button.
        TextButton* AddButton(const std::string &name) override;
        /// Adds a new button to this ContextMenu.
        /// @param name The label to apply to the new button.
        /// @param callback The function to run when the button is clicked.
        TextButton* AddButton(const std::string& name, const std::function<void()> &callback) override;
        /// Adds a separator to this ContextMenu.
        /// @param size The height of this separator.
        Separator* AddSeparator(const UDim& size = 5_px) override;
        /// Adds a button, which opens a submenu, to this ContextMenu. Only the submenu is returned.
        ContextMenu* AddSubmenu(const std::string& name) override;

        /// Sets the text size of the buttons and sub-menus managed by this object.
        void TextSize(int value) override;

        /// Sets the text color of the buttons and sub-menus managed by this object.
        void TextColor(const Color4& color) override;

        /// Sets the font of the buttons and sub-menus managed by this object.
        void Font(const JGL::Font& value) override;


        /// Sets whether the widget will automatically hide itself when the mouse leaves its bounding box.
        void CloseOnHoverLoss(bool value);

        /// @return Whether the widget will automatically hide itself when the mouse leaves its bounding box.
        [[nodiscard]] bool CloseOnHoverLoss() const;


        void Update(float delta) override;

    protected:
        VerticalListLayout* layout;
        bool close_on_hover_loss = true;
        float aggregate_height = 0;
        float estimated_width = 0;
        float mouse_outside_tests = 0;

        // TODO: This is handled the same way as in UtilityBar, proposed to abstract it away into NestableMenu.
        std::vector<TextButton*> buttons;
        std::vector<ContextMenu*> submenus;

    private:

    };
}
