/// Josh's User Interface Library
/// A C++20 Library for creating, styling, and rendering of a UI/UX widgets.
/// Developed and Maintained by Josh O'Leary @ Redacted Software.
/// Special Thanks to William Tomasine II and Maxine Hayes.
/// (c) 2024 Redacted Software
/// This work is dedicated to the public domain.

/// @file ScrollingRect.hpp
/// @desc Scrolling Rectangle Widget
/// @edit 2024-10-11

#pragma once
#include <JUI/Widgets/Rect.hpp>
namespace JUI {
    class ScrollingRect;
}

// TODO: Fix direction of ScrollBar itself
// TODO: MouseWheel scroll, and ScrollBar clickable
// TODO: Clamp range of ScrollBar

/// A Rectangle Widget which has a larger renderable area than the visible area.
/// This allows user-controlled scrolling of the viewable content.
class JUI::ScrollingRect : public Rect {
public:
    ~ScrollingRect() override;

    /// The default constructor initializes all child elements and members to reasonable defaults.
    ScrollingRect();

    /// Constructs a ScrollingRect by specifying the parent widget to bind it to.
    explicit ScrollingRect(Widget* parent);
public:
    JGL::RenderTarget* GetCanvas();
    [[nodiscard]] Vector2i CanvasSize() const;
    [[nodiscard]] float ScrollPos() const { return scroll; }
    [[nodiscard]] Vector2 CanvasPosition() const;
    // TODO scrolling in either direction. Assuming vertical scroll for now.
    Vector2 CanvasAbsolutePosition() const;
    Vector2 CanvasAbsoluteSize() const;
public:
    void ScrollPos(float pos) { scroll = pos; }
    void CanvasSize(const Vector2i& new_size);


    /// Sets the scroll amount of this ScrollingRect.
    /// The value will be clamped between ScrollMinimum() and ScrollMaximum().
    /// @param value The new scrolling amount to apply.
    /// @see ScrollMinimum(), ScrollMaximum().
    void SetScrollAmount(float value);

    /// Scrolls this ScrollingRect by the specified amount.
    /// @param amount The quantity of pixels by which to scroll the object.
    /// @see SetScrollAmount()
    float Scroll(float amount);

    /// @return The amount of pixels this scrolling rect is currently scrolled by.
    float GetScrollAmount() const;

    [[nodiscard]] bool CanScroll() const;
    /// The lower-bound of how far up the scroll can be.
    [[nodiscard]] float ScrollMinimum() const;
    /// The upper-bound of how far down the scroll can go.
    [[nodiscard]] float ScrollMaximum() const;

    void SetAutoAdjustScrollAmount(bool value);
    bool GetAutoAdjustScrollAmount() const;

    void ScrollToTop();

    void ScrollToBottom();

    // TODO: Figure out how to automatically adjust this as child widgets are added, removed, and changed in size.
    float canvas_height = 0;

public:
    void InnerDraw() override;
    void Draw() override;
    void Update(float delta) override;

    bool ObserveKeyInput(Key key, bool pressed) override;
    bool ObserveMouseWheel(int mwheel) override;
protected:
    bool vertical_scrollbar_enabled = true;
    bool horizontal_scrollbar_enabled = true;
    bool scrollbar_visible = true;
    float scrollbar_width = 12;
    Color4 scrollbar_color = Colors::Whites::Azure;
    float scroll = 0;
    float scroll_amount = 10;
    JGL::RenderTarget* canvas = nullptr;
    bool auto_adjust_scroll_amount = false;

protected:
    /* This isn't public because nothing should ever
     * have to do this from the outside. -Redacted */
    void RecomputeRenderTarget();

    void DrawVerticalScrollbar();
};