#include <JUI/Base/Widget.hpp>
#include <jlog/Logger.hpp>
#include <J3ML/Geometry/AABB2D.hpp>

using namespace JUI;

Widget::Widget() {
    name = "Widget";
    children = std::vector<Widget *>();
}
Widget::Widget(Widget* parent) : Widget() {
    this->Parent(parent);
 }

 void Widget::Parent(Widget* newParent) {
    // hold a reference to this so it doesn't get collected as we're working.
    Widget* oldParent = this->parent;

    // New parent is old parent, do nothing, maybe raise a warning.
    if (newParent == oldParent)
        return;

    // Don't allow for an instance to be parented to itself
    if (this == newParent)
        throw std::runtime_error("Cannot parent a widget to itself.");
    // You're trying to break the linearity of the object hierarchy by creating a circle in the graph.
    if (this->IsAncestorOf(newParent))
        throw std::runtime_error("Cannot create circular reference.");

    for (Widget* ancestor: this->GetAncestors()) {
        if (oldParent && !ancestor->IsAncestorOf(newParent) && newParent != ancestor) {
            ancestor->DescendantRemoved(this);
            for (Widget* descendant : this->GetDescendants())
                ancestor->DescendantRemoved(descendant);
        }
    }

    // Remove ourselves from our parent (if we have one)
    if (this->parent)
        std::erase(this->parent->children, this);

    // Update our old parent to the new one
    this->parent = newParent;

    // If our parent is set to nullptr, we can't update it's vector of children
    if (!newParent) return;

    // Add ourselves to our new parent's list of children
    newParent->children.emplace_back(this);
    //newParent->ChildAdded(this);

    for (Widget* ancestor : this->GetAncestors()) {
        if (!oldParent || (!oldParent->IsDescendantOf(newParent) && oldParent != ancestor)) {
            // Don't fire unless an instance is actually a new descendant
            ancestor->DescendantAdded(this);
            for (Widget* descendant: this->GetDescendants())
                ancestor->DescendantAdded(descendant);
        }
    }
}

bool Widget::IsDescendantOf(Widget* ancestor) {
    if (ancestor == nullptr) return false;

    Widget* instance = this;
    while (instance->GetParent()) {
        instance = instance->GetParent();
        if (instance == ancestor)
            return true;
    }
    return false;
}

bool Widget::IsAncestorOf(Widget* descendant) {
    if (descendant == nullptr) return false;
    Widget* instance = descendant;
    while (instance->GetParent()) {
        instance = instance->GetParent();
        if (instance == this)
            return true;
    }
    return false;
}

Vector2 Widget::AnchorPoint() const {
    return anchor_point;
}

UDim Widget::PaddingLeft() const { return pad_left;}
UDim Widget::PaddingTop() const { return pad_top;}
UDim Widget::PaddingRight() const { return pad_right;}
UDim Widget::PaddingBottom() const { return pad_bottom;}

void Widget::PaddingLeft  (const UDim& value)   { this->pad_left   = value;   }
void Widget::PaddingTop   (const UDim& pad_top)    { this->pad_top    = pad_top;    }
void Widget::PaddingRight (const UDim& pad_right)  { this->pad_right  = pad_right;  }
void Widget::PaddingBottom(const UDim& pad_bottom) { this->pad_bottom = pad_bottom; }

void Widget::Padding(const UDim &left, const UDim &top, const UDim &right, const UDim &bottom) {
    PaddingLeft(left);
    PaddingTop(top);
    PaddingRight(right);
    PaddingBottom(bottom);
}

void Widget::Padding(const UDim &padding) {
    Padding(padding, padding, padding, padding);
}

UDim Widget::MarginLeft  () const { return margin_left;}
UDim Widget::MarginTop   () const { return margin_top;}
UDim Widget::MarginRight () const { return margin_right;}
UDim Widget::MarginBottom() const { return margin_bottom;}

void Widget::MarginLeft(const UDim& ml) { margin_left = ml;}
void Widget::MarginTop(const UDim& mt) { margin_top = mt;}
void Widget::MarginRight(const UDim& mr) { margin_right = mr;}
void Widget::MarginBottom(const UDim& mb) { margin_bottom = mb;}

void Widget::Margin(const UDim &left, const UDim &top, const UDim &right, const UDim &bottom) {
    MarginLeft(left);
    MarginTop(top);
    MarginRight(right);
    MarginBottom(bottom);
}

void Widget::Margin(const UDim &margin) {
    Margin(margin, margin, margin, margin);
}

std::string Widget::Name() const { return name; }
void Widget::Name(const std::string& new_name) { name = new_name;}

bool Widget::IsVisible() const { return visible; }
void Widget::Visible(bool enabled) { visible = enabled;}

Widget* Widget::GetParent() const { return parent; }

UDim2 Widget::Position() const { return position; }
void Widget::Position(const UDim2& pos) { position = pos; }

UDim2 Widget::Size() const { return size;}
void Widget::Size(const UDim2& s) { size = s; }

float Widget::GetAbsoluteRotation() const {
    // TODO: implement rotation correctly!!
    return rotation;
}

Vector2 Widget::GetAbsoluteCentroid() const {
    return GetAbsolutePosition() + (GetAbsoluteSize() / 2.f);
}

std::vector<Widget *> Widget::GetChildren() { return children; }

std::vector<Widget *> Widget::GetDescendants() {
    std::vector<Widget*> descendants;
    for (auto& child: this->GetChildren()) {
        descendants.push_back(child);
        std::vector<Widget*>  recursiveDescendants = child->GetDescendants();
        descendants.insert(descendants.end(), recursiveDescendants.begin(), recursiveDescendants.end());
    }
    return descendants;
}

std::vector<Widget*> Widget::GetAncestors() {
    std::vector<Widget*> ancestors;
    for (Widget *ancestor = this->parent; ancestor; ancestor = ancestor->parent)
        ancestors.push_back(ancestor);
    return ancestors;
}

    /*float Widget::GetAbsoluteMarginLeft() {}
    float Widget::GetAbsoluteMarginRight() {}
    float Widget::GetAbsoluteMarginTop() {}
    float Widget::GetAbsoluteMarginBottom() {}
    float Widget::GetAbsolutePaddingLeft() {}
    float Widget::GetAbsolutePaddingRight() {}
    float Widget::GetAbsolutePaddingTop() {}
    float Widget::GetAbsolutePaddingBottom() {}*/



Vector2 Widget::GetAbsolutePaddingTopLeft() const {
    auto parent_abs_size = this->GetParent()->GetAbsoluteSize();
    UDim2 pad_topleft = {parent->PaddingLeft(), parent->PaddingTop()};
    return ComputeElementPadding(parent_abs_size, pad_topleft);
}

Vector2 Widget::GetAbsoluteMarginTopLeft() {
    // TODO: Implement correctly.
    return {0,0};
}

Vector2 Widget::GetAbsolutePaddingBottomRight() const {
    // Returns the amount by which the widget size should be shrunk.
    // This combines the Left + Right for horizontal,
    // and Top + Bottom for vertical padding.

    UDim padding_h = parent->PaddingLeft() + parent->PaddingRight();
    UDim padding_v = parent->PaddingTop() + parent->PaddingBottom();
    UDim2 pad_bottom_left = {padding_h, padding_v};

    return ComputeElementPadding(parent->GetAbsoluteSize(), pad_bottom_left);
}

Vector2 Widget::GetAbsoluteMarginBottomRight() {
    // TODO: Implement correctly.
    return {0,0};
}


Vector2 Widget::GetAbsolutePosition() const {

    if (this->parent == nullptr)
        return {0,0};

    auto child_pos_scale = this->Position().GetScale();
    auto child_pos_pixels = this->Position().GetPixels();
    auto parent_abs_size = this->GetParent()->GetAbsoluteSize();
    auto parent_abs_pos = this->GetParent()->GetAbsolutePosition();

    auto abs_size = GetAbsoluteSize();

    auto anchor_offset = abs_size * AnchorPoint();

    Vector2 padding_offset = GetAbsolutePaddingTopLeft();

    Vector2 absolute_position =
            parent_abs_pos + child_pos_pixels + (parent_abs_size * child_pos_scale) + padding_offset;

    return absolute_position - anchor_offset;
}

Vector2 Widget::GetAbsoluteSize() const {

    if (this->GetParent() == nullptr)
        return {0,0};

    Vector2 child_size_scale = this->Size().GetScale();
    Vector2 child_size_px = this->Size().GetPixels();

    Vector2 parent_abs_size = this->GetParent()->GetAbsoluteSize();

    Vector2 pad_size_reduction = GetAbsolutePaddingBottomRight();

    Vector2 abs_size = child_size_px + (parent_abs_size * child_size_scale) - pad_size_reduction;

    return abs_size;
}


std::optional<Widget*> Widget::FindFirstChild(const std::string& search_name) {
    for (auto& child : children) {
        if (child->Name() == search_name)
            return child;
    }
    return std::nullopt;
}


void Widget::Draw() {
    if (!visible)
        return;

    PreDraw();
    InnerDraw();
    PostDraw();
    DrawChildWidgets();
}

void Widget::UpdateTweens(float elapsed) {
    for (Tween* t: tweens)
        t->Update(elapsed);

    // TODO: Remove tweens if "dead"

    tweens.erase(std::remove_if(tweens.begin(), tweens.end(), [](Tween* t){
        if (t->HasCompleted())
        {
            return true;
        }
        return false;
    }), tweens.end());
}

void Widget::Update(float delta) {

    is_mouse_inside = ComputeIsMouseInside();

    UpdateChildWidgets(delta);
    UpdateTweens(delta);
}

struct {
    bool operator()(Widget* a, Widget* b) const { return a->ZIndex() < b->ZIndex();}
}

zIndexSort;

void Widget::DrawChildWidgets() {
    if (z_dirty)
        z_dirty = false,
        std::sort(children.begin(), children.end(), zIndexSort);

    for (auto* child : children)
        child->Draw();
}

void Widget::UpdateChildWidgets(float delta) {
    for (auto* child : children)
        child->Update(delta);
}

Widget* Widget::GetFamilyTreeRoot() const {
    // This is stupid, Fix ASAP!
    auto parent = this->GetParent();
    if (parent->GetParent() == nullptr)
        return parent;
    return parent->GetFamilyTreeRoot();
}


bool Widget::ComputeIsMouseInside() const {
    float x = last_known_mouse_pos.x;
    float y = last_known_mouse_pos.y;
    auto pos = GetAbsolutePosition();
    auto size = GetAbsoluteSize();

    if (x > pos.x && y > pos.y && x < pos.x + size.x && y < pos.y + size.y)
        return true;

    return false;
}

bool Widget::IsMouseInside() const {
    return is_mouse_inside;

    /*float x = last_known_mouse_pos.x;
    float y = last_known_mouse_pos.y;
    auto pos = GetAbsolutePosition();
    auto size = GetAbsoluteSize();

    if (x > pos.x && y > pos.y && x < pos.x + size.x && y < pos.y + size.y)
        return true;

    return false;*/
}

bool Widget::IsMouseInsideChildren(bool include_this) const {
    if (include_this)
        if (IsMouseInside())
            return true;

    for (auto* child : children) {
        if (child->IsMouseInside())
            return true;
    }
    return false;
}

bool Widget::IsMouseInsideDescendants(bool include_this) const {
    if (include_this)
        if (IsMouseInside())
            return true;
    for (auto* child : children) {
        if (child->IsMouseInsideDescendants(false))
            return true;
    }
    return false;
}

bool Widget::ObserveMouseInput(MouseButton btn, bool pressed) {
    mbtn = btn;
    mb_state = pressed;

    for (Widget* child : children) {
        if (child->ObserveMouseInput(btn, pressed))
            return true;
    }
    return false;
}

bool Widget::ObserveMouseMovement(const Vector2 &latest_known_pos) {
    last_known_mouse_pos = latest_known_pos;

    for (Widget* child: children)
        if (child->ObserveMouseMovement(latest_known_pos))
            return true;

    return false;
}

bool Widget::ObserveKeyInput(Key key, bool pressed) {
    for (Widget* child : children)
        if (child->ObserveKeyInput(key, pressed))
            return true;

    return false;
}

bool Widget::ObserveMouseWheel(int mwheel) {
    // TODO: Log that this is a stub.
    for (auto& child : children) {
        if (child->ObserveMouseWheel(mwheel))
            return true;
    }
    return false;
}

bool Widget::IsAncestorOf(Widget *descendant) const {
    if (descendant == nullptr)
        return false;

    Widget* instance = descendant;
    while (instance->GetParent()) {
        instance = instance->GetParent();
        if (instance == this)
            return true;
    }
    return false;
}

bool Widget::IsDescendantOf(Widget *ancestor) const {
    if (ancestor == nullptr)
        return false;
    const Widget *instance = this;
    while (instance->GetParent()) {
        instance = instance->GetParent();
        if (instance == ancestor)
            return true;
    }
    return false;
}

Widget *Widget::Add(Widget *newChild) {
    newChild->Parent(this);
    return newChild;
}

void Widget::AnchorPoint(const Vector2& point) {
    anchor_point = point;
}

int Widget::ZIndex() const { return zindex;}

void Widget::ZIndex(int new_zindex) {
zindex = new_zindex;
    if (parent)
        parent->z_dirty = true;
}

AABB2D Widget::GetActualRenderBounds() const {
    return {GetAbsolutePosition(), GetAbsoluteSize()};
}

void Widget::SetViewportSize(const Vector2 &vps) {
    viewport_size = vps;

    for(auto* child : children)
        child->SetViewportSize(viewport_size);
}

Vector2 Widget::ComputeElementPadding(const Vector2 &size, const UDim2 &padding) const {
    float pad_x = padding.X.Pixels + (padding.X.Scale * size.x);
    float pad_y = padding.Y.Pixels + (padding.Y.Scale * size.y);

    return {pad_x, pad_y};
}

AABB2D Widget::AbsoluteBounds() const {
    return {GetAbsolutePosition(), GetAbsoluteSize()};
}

Tween* Widget::TweenPosition(const UDim2& goal, TweenInfo params) {
    UDim2 start_position = this->Position();
    TweenTickFunc updateTillGoalReached = [this, start_position, goal] (float elapsed, float progress) mutable {
        UDim2 pos = start_position.Lerp(goal, progress);
        Position(pos);
    };

    Tween* t = new Tween(updateTillGoalReached, params);
    tweens.push_back(t);
    return t;
}

Tween* Widget::TweenSize(const UDim2& goal, TweenInfo params) {
    UDim2 start_size = this->Size();
    TweenTickFunc updateTillGoalReached = [this, start_size, goal] (float elapsed, float progress) mutable {
        UDim2 step_size = start_size.Lerp(goal, progress);
        Size(step_size);
    };

    Tween* t = new Tween(updateTillGoalReached, params);
    tweens.push_back(t);
    return t;
}

Tween* Widget::TweenPaddingLeft(const JUI::UDim& goal, JUI::TweenInfo info) {
    //UDum2
    UDim start_padding = this->PaddingLeft();
    TweenTickFunc updateTillGoalReached = [this, start_padding, goal] (float elapsed, float progress) mutable {
        UDim step_padding = start_padding.Lerp(goal, progress);
        PaddingLeft(step_padding);
    };

    Tween* t = new Tween(updateTillGoalReached, info);
    tweens.push_back(t);

    return t;
}

Tween* Widget::TweenPaddingRight(const JUI::UDim& goal, JUI::TweenInfo info) {
    //UDum2
    UDim start_padding = this->PaddingRight();
    TweenTickFunc updateTillGoalReached = [this, start_padding, goal] (float elapsed, float progress) mutable {
        UDim step_padding = start_padding.Lerp(goal, progress);
        PaddingRight(step_padding);
    };

    Tween* t = new Tween(updateTillGoalReached, info);
    tweens.push_back(t);
    return t;
}

Tween* Widget::TweenPaddingTop(const JUI::UDim& goal, JUI::TweenInfo info) {
    //UDum2
    UDim start_padding = this->PaddingTop();
    TweenTickFunc updateTillGoalReached = [this, start_padding, goal] (float elapsed, float progress) mutable {
        UDim step_padding = start_padding.Lerp(goal, progress);
        PaddingTop(step_padding);
    };

    Tween* t = new Tween(updateTillGoalReached, info);
    tweens.push_back(t);
    return t;
}

Tween* Widget::TweenPaddingBottom(const JUI::UDim& goal, JUI::TweenInfo info) {
     //UDum2
     UDim start_padding = this->PaddingBottom();
     TweenTickFunc updateTillGoalReached = [this, start_padding, goal] (float elapsed, float progress) mutable {
         UDim step_padding = start_padding.Lerp(goal, progress);
         PaddingBottom(step_padding);
     };

     Tween* t = new Tween(updateTillGoalReached, info);
     tweens.push_back(t);
     return t;
}

    Tween *Widget::TweenPadding(const JUI::UDim &gl, const JUI::UDim &gt, const JUI::UDim &gr,
                                const JUI::UDim &gb, JUI::TweenInfo info) {
    UDim spl = this->PaddingLeft();
    UDim spr = this->PaddingRight();
    UDim spt = this->PaddingTop();
    UDim spb = this->PaddingBottom();

    TweenTickFunc updateTillGoalReached = [this, spl, spr, spt, spb, gl, gt, gr, gb] (float elapsed, float progress) mutable {
        UDim step_l = spl.Lerp(gl, progress);
        UDim step_r = spr.Lerp(gr, progress);
        UDim step_b = spb.Lerp(gb, progress);
        UDim step_t = spt.Lerp(gt, progress);
        Padding(step_l, step_r, step_b, step_t);
    };

    Tween* t = new Tween(updateTillGoalReached, info);
    tweens.push_back(t);
    return t;
}

Tween *Widget::TweenPadding(const JUI::UDim &goal, JUI::TweenInfo info) {
    UDim start_padding = this->PaddingTop(); // TODO: Determine which side is most appropriate as start_point.
    TweenTickFunc updateTillGoalReached = [this, start_padding, goal] (float elapsed, float progress) mutable {
        UDim step_padding = start_padding.Lerp(goal, progress);
        Padding(step_padding);
    };

    Tween* t = new Tween(updateTillGoalReached, info);
    tweens.push_back(t);
    return t;
}

    Tween *Widget::TweenMargin(const JUI::UDim &gl, const JUI::UDim &gt, const JUI::UDim &gr,
                                const JUI::UDim &gb, JUI::TweenInfo info) {
    UDim spl = this->MarginLeft();
    UDim spr = this->MarginRight();
    UDim spt = this->MarginTop();
    UDim spb = this->MarginBottom();

    TweenTickFunc updateTillGoalReached = [this, spl, spr, spt, spb, gl, gt, gr, gb] (float elapsed, float progress) mutable {
        UDim step_l = spl.Lerp(gl, progress);
        UDim step_r = spr.Lerp(gr, progress);
        UDim step_b = spb.Lerp(gb, progress);
        UDim step_t = spt.Lerp(gt, progress);
        Margin(step_l, step_r, step_b, step_t);
    };

    Tween* t = new Tween(updateTillGoalReached, info);
    tweens.push_back(t);
    return t;
}

Tween *Widget::TweenMargin(const JUI::UDim &goal, JUI::TweenInfo info) {
    UDim start_Margin = this->MarginTop(); // TODO: Determine which side is most appropriate as start_point.
    TweenTickFunc updateTillGoalReached = [this, start_Margin, goal] (float elapsed, float progress) mutable {
        UDim step_Margin = start_Margin.Lerp(goal, progress);
        Margin(step_Margin);
    };

    Tween* t = new Tween(updateTillGoalReached, info);
    tweens.push_back(t);
    return t;
}

Tween* Widget::TweenMarginLeft(const JUI::UDim& goal, JUI::TweenInfo info) {
    UDim start = this->MarginLeft();
    TweenTickFunc updateTillGoalReached = [this, start, goal] (float elapsed, float progress) mutable {
        UDim step = start.Lerp(goal, progress);
        MarginLeft(step);
    };

    Tween* t = new Tween(updateTillGoalReached, info);
    tweens.push_back(t);
    return t;
}

Tween* Widget::TweenMarginRight(const JUI::UDim& goal, JUI::TweenInfo info) {
    UDim start = this->MarginRight();
    TweenTickFunc updateTillGoalReached = [this, start, goal] (float elapsed, float progress) mutable {
        UDim step = start.Lerp(goal, progress);
        MarginRight(step);
    };

    Tween* t = new Tween(updateTillGoalReached, info);
    tweens.push_back(t);
    return t;
}

Tween* Widget::TweenMarginTop(const JUI::UDim& goal, JUI::TweenInfo info) {
    UDim start = this->MarginTop();
    TweenTickFunc updateTillGoalReached = [this, start, goal] (float elapsed, float progress) mutable {
        UDim step = start.Lerp(goal, progress);
        MarginTop(step);
    };

    Tween* t = new Tween(updateTillGoalReached, info);
    tweens.push_back(t);
    return t;
}

Tween* Widget::TweenMarginBottom(const JUI::UDim& goal, JUI::TweenInfo info) {
    UDim start = this->MarginBottom();
    TweenTickFunc updateTillGoalReached = [this, start, goal] (float elapsed, float progress) mutable {
        UDim step = start.Lerp(goal, progress);
        MarginBottom(step);
    };
    
    Tween* t = new Tween(updateTillGoalReached, info);
    tweens.push_back(t);
    return t;
}

Tween *Widget::TweenAnchorPoint(const Vector2 &goal, TweenInfo info) {
    Vector2 start = this->AnchorPoint();

    TweenTickFunc updateTillGoalReached = [this, start, goal] (float elapsed, float progress) mutable {
        Vector2 step = start.Lerp(goal, progress);

        AnchorPoint(step);
    };

    Tween* t = new Tween(updateTillGoalReached, info);
    tweens.push_back(t);
    return t;
}



float Widget::ComputeElementPadding(float size, const UDim &padding) {
    return padding.Pixels + (padding.Scale * size);
}

Widget* Widget::GetHighestNonRootAncestor() {
    if (!parent)
        return nullptr;

    Widget* previous = this;
    Widget* current = parent;

    while (current->parent) {
        previous = current;
        current = current->parent;
    }

    return previous;
}


