#include <JUI/Widgets/ScrollingRect.hpp>


using namespace JUI;

void ScrollingRect::RecomputeRenderTarget() {
    InnerDraw();
}


void ScrollingRect::DrawVerticalScrollbar() {

    // TODO: Ratio for size is fine, but the computation for the scrollbar position is just a raw pixel-offset
    // when it needs to be a size_ratio of the amount scrolled.

    float size_ratio = canvas_height / GetAbsoluteSize().y;
    float pos_ratio = scroll / canvas_height;


    float travel = (GetAbsoluteSize().y * pos_ratio);

    Vector2 scrollbar_base_pos = GetAbsolutePosition() + Vector2(GetAbsoluteSize().x - scrollbar_width, 0);
    Vector2 scrollbar_base_size = {scrollbar_width, GetAbsoluteSize().y};

    // Draw box for scrollbar to fit into.
    J2D::FillRect(Colors::LightGray, scrollbar_base_pos, scrollbar_base_size);


    // Compute change in size and position that is appropriate.
    Vector2 scrollbar_pos = scrollbar_base_pos + Vector2(0, travel);
    Vector2 scrollbar_size = {scrollbar_base_size.x, (scrollbar_base_size.y / size_ratio)};

    J2D::FillRoundedRect(scrollbar_color, scrollbar_pos, scrollbar_size, 4);

    // Draw little square box at the point where the vertical and horizontal scrollbars would meet.
    //Vector2 lil_box_size = {scrollbar_width, scrollbar_width};
    //Vector2 lil_box_pos = GetAbsolutePosition() + GetAbsoluteSize() - lil_box_size;
    //J2D::FillRect(Colors::DarkGray, lil_box_pos, lil_box_size);
}

void ScrollingRect::Draw() {

    // TODO: When `scroll` is 0, we are at the **bottom** of the render target.
    // TODO: Flip this around so 0 is at the top
    // TODO: Also flip the scroll-bar around.

    // TODO: Only recompute when something has changed.
    // Q: How do we know when something has changed?
    RecomputeRenderTarget();

    J2D::DrawPartialRenderTarget(canvas, GetAbsolutePosition(), {0, 0}, Vector2(GetAbsoluteSize().x, GetAbsoluteSize().y));

    bool canvas_larger_than_widget = canvas_height > GetAbsoluteSize().y;

    if (vertical_scrollbar_enabled && canvas_larger_than_widget) {
        DrawVerticalScrollbar();
    }
    //J2D::DrawString(Colors::Black, std::format("scroll {}, canvas {}", scroll, canvas_height), GetAbsolutePosition().x, GetAbsolutePosition().y, 1, 12);
}

Vector2 ScrollingRect::CanvasPosition() const {
    return {GetAbsolutePosition().x, GetAbsolutePosition().y - scroll};
}

// Wouldn't inner draw actually render everything onto the RenderTarget?
void ScrollingRect::InnerDraw() {

    J2D::Begin(canvas, nullptr, true);
        auto saved_pos = position;
        auto saved_size = size;

        // Kind of a hack to get the child widgets onto the rendertarget at the correct position.
        // We basically set the position for just long enough to lie to the child widgets.
        // The amount we offset by is the absolute position.
        // If you work the formulas out, the child widgets will believe their parent's AbsolutePosition is {0,0}.

        Position(UDim2::FromPixels(-GetAbsolutePosition().x, -GetAbsolutePosition().y + scroll));

        // TODO: Do something with this fake canvas-size?
        // TODO: How frequently can we update the RenderTarget size.
        Size({200_percent, 500_percent});
        Rect::InnerDraw();
        DrawChildWidgets();

        //J2D::DrawPoint(Colors::Blue, {1,1}, 2);
        //J2D::DrawPoint(Colors::Blue, Vector2(GetAbsoluteSize()), 2);

        // Set our position back once we're done.
        Size(saved_size);
        Position(saved_pos);
    J2D::End();
}

const Vector2i default_initialize_canvas_size {1024, 4096};

ScrollingRect::~ScrollingRect() { delete canvas; }

ScrollingRect::ScrollingRect() : Rect(), canvas(new RenderTarget(default_initialize_canvas_size)) {
    Name("ScrollingRect");
    //bool success = canvas->SetMSAAEnabled(JGL::SampleRate::X8);
}

ScrollingRect::ScrollingRect(Widget *parent) : ScrollingRect() {
    Parent(parent);
    //bool success = canvas->SetMSAAEnabled(JGL::MSAA_SAMPLE_RATE::MSAA_8X);
}

JGL::RenderTarget *ScrollingRect::GetCanvas() { return canvas; }

Vector2i ScrollingRect::CanvasSize() const { return canvas->GetDimensions(); }

void ScrollingRect::Update(float delta) {
    //scroll += delta*5;
    //canvas->Resize(Vector2i(GetAbsoluteSize().x, GetAbsoluteSize().y));
    Rect::Update(delta);
}

bool ScrollingRect::ObserveKeyInput(Key key, bool pressed) {
    if (Widget::ObserveKeyInput(key, pressed))
        return true;

    if (IsMouseInside()) {
        if (key == Keys::UpArrow && pressed)
        {
            Scroll(-scroll_amount);
            return true;
        }
        if (key == Keys::DownArrow && pressed)
        {
            Scroll(scroll_amount);
            return true;
        }

        if (key == Keys::PageDown && pressed) {
            ScrollToBottom();
            return true;
        }

        if (key == Keys::PageUp && pressed) {
            ScrollToTop();
            return true;
        }
    }
    // TODO: Forego this in favor of mouse scrolling and clicking.

    return false;
}

bool ScrollingRect::CanScroll() const {
    return canvas_height > GetAbsoluteSize().y;
}

void ScrollingRect::SetScrollAmount(float value) {
    scroll = Math::Clamp(value, ScrollMinimum(), ScrollMaximum());
}

float ScrollingRect::Scroll(float amount) {
    SetScrollAmount(scroll + amount);
    return scroll;
}

float ScrollingRect::ScrollMaximum() const {
    return Math::Max(0.f, canvas_height-GetAbsoluteSize().y);
}

void ScrollingRect::ScrollToTop() {
    SetScrollAmount(ScrollMinimum());
}

void ScrollingRect::ScrollToBottom() {
    SetScrollAmount(ScrollMaximum());
}

float ScrollingRect::ScrollMinimum() const { return 0; }

float ScrollingRect::GetScrollAmount() const { return scroll; }

bool ScrollingRect::ObserveMouseWheel(int mwheel) {
    if (Widget::ObserveMouseWheel(mwheel))
        return true;

    if (IsMouseInside()) {
        Scroll(mwheel*scroll_amount);
        return true;
    }
    return false;
}
