#include <JUI/Widgets/Slider.hpp>


namespace JUI
{

    template <typename T>
    struct Range {
        T min;
        T max;
    };

    template <typename T>
    T map(T value, T in_min, T in_max, T out_min, T out_max) {
        return (value - in_min) / (in_max - in_min) * (out_max - out_min) + out_min;
    }

    template <typename T>
    T map(T value, const Range<T>& in, const Range<T>& out) {
        return map(value, in.min, in.max, out.min, out.max);
    }

    template<typename T>
    T roundMultiple( T value, T multiple )
    {
        if (multiple == 0) return value;
        return static_cast<T>(std::round(static_cast<double>(value)/static_cast<double>(multiple))*static_cast<double>(multiple));
    }

    Slider::Slider(JUI::Widget *parent) : Slider()
    {
        this->Parent(parent);
    }

    void Slider::OnClick(const Vector2 &MousePos, const MouseButton &MouseButton) {
        Clickable::OnClick(MousePos, MouseButton);
        dragging = true;
    }

    void Slider::OnRelease(const Vector2 &MousePos, const MouseButton &MouseButton, bool MouseStillOver) {
        Clickable::OnRelease(MousePos, MouseButton, MouseStillOver);
        dragging = false;
    }

    void Slider::OnHover(const Vector2 &MousePos) {
        Hoverable::OnHover(MousePos);
    }

    void Slider::OnExit(const Vector2 &MousePos) {
        Hoverable::OnExit(MousePos);
    }

    void Slider::Update(float delta) {

        Hoverable::Update(Rect::IsMouseInside(), delta);

        // TODO: This is duplicated between here and Window.cpp
        // Will be abstracted away into Clickable shortly.
        if (IsHovered() && mb_state && !prev_mb_state)
        {
            OnClick(last_known_mouse_pos, mbtn);
        }

        if (IsClicked() && !mb_state)
        {
            OnRelease(last_known_mouse_pos, mbtn, IsHovered());
        }

        prev_mb_state = mb_state;

        if (dragging)
        {
            Vector2 mouse = last_known_mouse_pos;

            float scrubber_width = GetScrubberAbsoluteSize().x;

            float slider_abs_left = this->GetAbsolutePosition().x;
            float slider_abs_right = (this->GetAbsolutePosition().x + this->GetAbsoluteSize().x) - scrubber_width;
            float slider_total_width = this->GetAbsoluteSize().x - scrubber_width;

            float scrubberP = Math::Clamp(mouse.x, slider_abs_left, slider_abs_right);

            float offset = scrubberP - slider_abs_left;

            percentage = offset / slider_total_width;

            float translated = Math::Clamp(roundMultiple(percentage * Range(), interval), 0.f, Range());

            float value = translated + minimum;

            if (current != value)
                ValueChanged(value);

            current = value;
        }

        // TODO there is probably a more elegant fix for this.
       // Fixes the slider not letting go - Redacted.
       Widget* target = parent ? GetHighestNonRootAncestor() : this;
       if (!target->IsMouseInside()) { dragging = false; mb_state = false; }

        Rect::Update(delta);
    }


    /// @note Calls GetScrubberAbsoluteSize() to compute box's offset.
    Vector2 Slider::GetScrubberAbsolutePosition() const {

        // TODO: Integrate padding.

        float x_offset = percentage * (GetAbsoluteSize().x - GetScrubberAbsoluteSize().x);

        float rect_half_height = (GetAbsoluteSize().y / 2.f);

        float y_offset = rect_half_height - (GetScrubberAbsoluteSize().y / 2.f);

        Vector2 scrub_rel_pos = {x_offset, y_offset};

        return GetAbsolutePosition() + scrub_rel_pos;

    }

    Vector2 Slider::GetScrubberAbsoluteSize() const {
        Vector2 abs_size = GetAbsoluteSize();
        Vector2 scrub_size_px = scrubber_size.GetPixels();
        Vector2 scrub_size_scale = scrubber_size.GetScale();

        // TODO: Implement Slider::ScrubberPadding();
        Vector2 padding = GetAbsolutePaddingBottomRight();

        return scrub_size_px + (abs_size * scrub_size_scale) - padding;
    }

    void Slider::InnerDraw() {
        Rect::InnerDraw();

        Vector2 pos = GetScrubberAbsolutePosition();
        Vector2 size = GetScrubberAbsoluteSize();

        RectBase::Draw(scrubber_color, border_color, pos, size, scrubber_rounding, border_width, border_mode, corner_mode);
    }

    void Slider::Draw() {
        Rect::Draw();
    }

    float Slider::Minimum() const { return minimum; }

    float Slider::Maximum() const { return maximum; }

    float Slider::Interval() const { return interval; }

    float Slider::CurrentValue() const {
        return map(percentage, {0, 1}, {minimum, maximum});
    }

    Color4 Slider::ScrubberColor() const { return scrubber_color; }

    float Slider::ScrubberWidth() const { return scrubber_size.X.Pixels; }

    UDim2 Slider::ScrubberSize() const {
        return scrubber_size;
    }

    float Slider::ScrubberRounding() const {
        return scrubber_rounding;
    }

    float Slider::Range() const { return maximum - minimum; }

    void Slider::Minimum(float min) { minimum = min; }

    void Slider::Maximum(float max) { maximum = max; }

    void Slider::Interval(float inter) { interval = inter; }

    void Slider::CurrentValue(float value) {
        percentage = map(value, {minimum, maximum}, {0.f, 1.f});
    }

    float Slider::Percentage() const { return percentage;}
    void Slider::Percentage(float value) { percentage = value; }

    void Slider::ScrubberColor(const Color4 &color) { scrubber_color = color;}

    void Slider::ScrubberWidth(float width) { scrubber_size.X.Pixels = width;}

    void Slider::ScrubberSize(const UDim2 &size) {
        scrubber_size = size;
    }

    void Slider::ScrubberRounding(float rounding) { scrubber_rounding = rounding;}

    Slider::Slider() {
        Name("Slider");
    }

    void Slider::SetDragging(bool value) {
        dragging = value;
    }

    bool Slider::Dragging() const { return dragging; }
}